import typing
from dataclasses import dataclass
from types import UnionType
from typing import Any

_implict_modules = ["builtins", "typing", "collections.abc", "types"]


@dataclass(frozen=True, slots=True)
class TypeId:
    """
    Represents information about a Type in the Dependency Injection framework.
    """

    type: type
    multi: bool

    @classmethod
    def from_type(cls, type_: Any) -> "TypeId":
        """
        Construct a TypeId from a given type.

        Args:
            type_: any type.

        Returns:
            The corresponding TypeId for that type.
        """
        if _is_multi_type(type_):
            inner_obj = typing.get_args(type_)[0]
            return TypeId(type=inner_obj, multi=True)
        else:
            return TypeId(type=type_, multi=False)

    def __str__(self) -> str:
        module = getattr(self.type, "__module__", None)
        out = f"{module}." if module and module not in _implict_modules else ""
        out += _args_to_str(self.type)
        if self.multi:
            out += "[]"
        return out

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, TypeId):
            return False
        return self.type == other.type and self.multi == other.multi


def _args_to_str(type_: Any) -> str:
    args = typing.get_args(type_)
    if args:
        arg_str = "Union[" if isinstance(type_, UnionType) else f"{type_.__name__}["
        for idx, arg in enumerate(args):
            if isinstance(arg, list):
                arg_str += "["
                for inner_idx, inner_arg in enumerate(arg):
                    arg_str += _args_to_str(inner_arg)
                    if inner_idx < len(arg) - 1:
                        arg_str += ", "
                arg_str += "]"
            elif typing.get_args(arg):
                arg_str += _args_to_str(arg)
            else:
                arg_str += getattr(arg, "__name__", str(arg))
            if idx < len(args) - 1:
                arg_str += ", "
        arg_str += "]"
    else:
        arg_str = type_.__name__
    return arg_str


def _is_multi_type(type_: Any) -> bool:
    """
    Discriminates a type to determine whether it is the return type of a multiprovider.
    """
    return typing.get_origin(type_) is list



import importlib
import sys
from pathlib import Path

import typer
from rich import print
from rich.panel import Panel
from typing_extensions import Never

from engin import Engin

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def print_error(msg: str) -> Never:
    print(
        Panel(
            title="Error",
            renderable=msg,
            title_align="left",
            border_style="red",
            highlight=True,
        )
    )
    raise typer.Exit(code=1)


COMMON_HELP = {
    "app": (
        "The import path of your Engin instance, in the form 'package:application'"
        ", e.g. 'app.main:engin'. If not provided, will try to use the `default-instance`"
        " value specified in your pyproject.toml"
    )
}


def _find_pyproject_toml() -> Path | None:
    """Find pyproject.toml file starting from current directory and walking up."""
    current_path = Path.cwd()

    for path in [current_path, *current_path.parents]:
        pyproject_path = path / "pyproject.toml"
        if pyproject_path.exists():
            return pyproject_path

    return None


def _get_default_engin_from_pyproject() -> str | None:
    """Get the default engin instance from pyproject.toml."""
    pyproject_path = _find_pyproject_toml()
    if not pyproject_path:
        return None

    try:
        with Path(pyproject_path).open("rb") as f:
            data = tomllib.load(f)

        tool_section = data.get("tool", {})
        engin_section = tool_section.get("engin", {})
        instance = engin_section.get("default-instance")

        if instance is None:
            return None

        if not isinstance(instance, str):
            print_error("value of `default-instance` is not a string")

        return instance

    except (OSError, tomllib.TOMLDecodeError):
        print_error("invalid toml detected")


NO_APP_FOUND_ERROR = (
    "App path not specified and no default instance specified in pyproject.toml"
)


def get_engin_instance(app: str | None = None) -> tuple[str, str, Engin]:
    """
    Get an Engin instance either from the provided value or from pyproject.toml.

    Args:
        app: Optional string in format 'module:attribute'. If not provided will lookup in
            pyproject.toml.

    Returns:
        Tuple of (module_name, engin_name, engin_instance)

    Raises:
        typer.Exit: If no app is provided and no default instance is specified in the user's
            pyproject.toml.
    """
    if app is None:
        app = _get_default_engin_from_pyproject()
        if app is None:
            print_error(NO_APP_FOUND_ERROR)

    try:
        module_name, engin_name = app.split(":", maxsplit=1)
    except ValueError:
        print_error("Expected an argument of the form 'module:attribute', e.g. 'myapp:engin'")

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        print_error(f"Unable to find module '{module_name}'")

    try:
        instance = getattr(module, engin_name)
    except AttributeError:
        print_error(f"Module '{module_name}' has no attribute '{engin_name}'")

    if not isinstance(instance, Engin):
        print_error(f"'{app}' is not an Engin instance")

    return module_name, engin_name, instance



from engin._assembler import Assembler
from engin._block import Block, invoke, provide
from engin._dependency import Entrypoint, Invoke, Provide, Supply
from engin._engin import Engin
from engin._lifecycle import Lifecycle
from engin._option import Option
from engin._supervisor import OnException, Supervisor
from engin._type_utils import TypeId

__all__ = [
    "Assembler",
    "Block",
    "Engin",
    "Entrypoint",
    "Invoke",
    "Lifecycle",
    "OnException",
    "Option",
    "Provide",
    "Supervisor",
    "Supply",
    "TypeId",
    "invoke",
    "provide",
]



import inspect
import logging
import typing
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from enum import Enum
from types import TracebackType
from typing import TypeAlias

import anyio
from anyio import get_cancelled_exc_class
from typing_extensions import assert_never

if typing.TYPE_CHECKING:
    from anyio.abc import TaskGroup

LOG = logging.getLogger("engin")

AsyncFunction: TypeAlias = Callable[[], Awaitable[None]]


class OnException(Enum):
    SHUTDOWN = 0
    """
    Cancel all other supervised tasks and shutdown the Engin.
    """

    RETRY = 1
    """
    Retry the task.
    """

    IGNORE = 2
    """
    The task will be not be retried and the engin will not be stopped, other tasks will
    continue to run.
    """


@dataclass(kw_only=True, slots=True, eq=False)
class _SupervisorTask:
    """
    Attributes:
        - factory: a coroutine function that can create the task.
        - on_exception: determines the behaviour when the task raises an exception.
        - complete: will be set to true if task stops for any reason except cancellation.
        - last_exception: the last exception raised by the task.
    """

    factory: AsyncFunction
    on_exception: OnException
    complete: bool = False
    last_exception: Exception | None = None
    shutdown_hook: AsyncFunction | None = None

    async def __call__(self) -> None:
        # loop to allow for restarting erroring tasks
        while True:
            try:
                await self.factory()
                self.complete = True
                return
            except get_cancelled_exc_class() as err:
                LOG.debug(f"supervised task '{self.name}' was cancelled", exc_info=err)
                raise
            except Exception as err:
                self.last_exception = err
                if self.on_exception == OnException.IGNORE:
                    LOG.warning(
                        f"supervisor task '{self.name}' raised {type(err).__name__} "
                        "which will be ignored",
                        exc_info=err,
                    )
                    self.complete = True
                    return
                if self.on_exception == OnException.RETRY:
                    LOG.warning(
                        f"supervisor task '{self.name}' raised {type(err).__name__} "
                        "which will be retried",
                        exc_info=err,
                    )
                    continue
                if self.on_exception == OnException.SHUTDOWN:
                    LOG.error(
                        f"supervisor task '{self.name}' raised {type(err).__name__}, "
                        "starting shutdown",
                        exc_info=err,
                    )
                    self.complete = True
                    raise get_cancelled_exc_class() from None
                assert_never(self.on_exception)

    @property
    def name(self) -> str:
        factory = self.factory
        if inspect.ismethod(factory):
            return f"{factory.__self__.__class__.__name__}.{factory.__func__.__name__}"
        if inspect.isclass(factory):
            return type(factory).__name__
        if inspect.isfunction(factory):
            return factory.__name__
        return str(factory)


class Supervisor:
    def __init__(self) -> None:
        self._tasks: list[_SupervisorTask] = []
        self._task_group: TaskGroup | None = None

    def supervise(
        self,
        func: AsyncFunction,
        *,
        on_exception: OnException = OnException.SHUTDOWN,
        shutdown_hook: AsyncFunction | None = None,
    ) -> None:
        self._tasks.append(
            _SupervisorTask(
                factory=func, on_exception=on_exception, shutdown_hook=shutdown_hook
            )
        )

    @property
    def empty(self) -> bool:
        return not self._tasks

    async def __aenter__(self) -> None:
        if not self._tasks:
            return

        self._task_group = await anyio.create_task_group().__aenter__()

        for task in self._tasks:
            LOG.info(f"supervising task: {task.name}")
            self._task_group.start_soon(task, name=task.name)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None:
        if self._task_group:
            for task in self._tasks:
                if task.shutdown_hook is not None:
                    LOG.debug(f"supervised task shutdown hook: {task.name}")
                    await task.shutdown_hook()
            if not self._task_group.cancel_scope.cancel_called:
                self._task_group.cancel_scope.cancel()
            await self._task_group.__aexit__(exc_type, exc_value, traceback)



import contextlib
import json
import socketserver
import threading
from http.server import BaseHTTPRequestHandler
from pathlib import Path
from time import sleep
from typing import Annotated, Any

import typer
from rich import print

from engin import Engin, Entrypoint, Invoke, TypeId
from engin._cli._common import COMMON_HELP, get_engin_instance
from engin._dependency import Dependency, Provide, Supply
from engin._graph import Node
from engin.extensions.asgi import ASGIEngin

try:
    from engin.extensions.fastapi import APIRouteDependency
except ImportError:
    APIRouteDependency = None  # type: ignore[assignment,misc]

cli = typer.Typer()


_APP_ORIGIN = ""


@cli.command(name="graph")
def serve_graph(
    app: Annotated[
        str | None,
        typer.Argument(help=COMMON_HELP["app"]),
    ] = None,
    port: Annotated[
        int,
        typer.Option("--port", "-p", help="Port to serve the graph on."),
    ] = 8123,
) -> None:
    """
    Creates a visualisation of your application's dependencies.

    Examples:

        1. `engin graph`
        2. `engin graph app.main:engin -p 8123`
    """
    module_name, _, instance = get_engin_instance(app)

    global _APP_ORIGIN
    _APP_ORIGIN = module_name.split(".", maxsplit=1)[0]

    nodes = instance.graph()

    # Generate JSON data for interactive graph
    graph_data = _generate_graph_data(nodes, instance)

    html = _GRAPH_HTML.replace("%%GRAPH_DATA%%", json.dumps(graph_data, indent=2)).encode(
        "utf8"
    )

    class Handler(BaseHTTPRequestHandler):
        def do_GET(self) -> None:
            self.send_response(200, "OK")
            self.send_header("Content-type", "html")
            self.end_headers()
            self.wfile.write(html)

        def log_message(self, format: str, *args: Any) -> None:
            return

    shutdown_event = threading.Event()

    def _shutdown_thread(server: socketserver.TCPServer) -> None:
        shutdown_event.wait()  # Wait until the shutdown event is set
        server.shutdown()  # This will cause server.serve_forever() to exit

    def _start_server() -> None:
        with socketserver.TCPServer(("localhost", port), Handler) as httpd:
            print(f"Serving dependency graph on http://localhost:{port}")
            shutdown_thread = threading.Thread(target=_shutdown_thread, args=(httpd,))
            shutdown_thread.start()  # Start shutdown event listener thread
            httpd.serve_forever()  # Will only return when shutdown() is called
            shutdown_thread.join()  # Ensure shutdown thread has finished

    server_thread = threading.Thread(target=_start_server)
    server_thread.start()

    with contextlib.suppress(KeyboardInterrupt):
        wait_for_interrupt()

    print("Exiting the server...")
    shutdown_event.set()  # notify the shutdown
    server_thread.join()  # Wait for the server thread to finish


def wait_for_interrupt() -> None:
    sleep(10000)


def _generate_graph_data(nodes: list[Node], instance: Engin) -> dict[str, Any]:
    """Generate JSON data structure for interactive graph rendering."""
    all_deps = set()
    for node in nodes:
        all_deps.add(node.node)
        if node.parent:
            all_deps.add(node.parent)

    # Generate node data
    node_data = []
    for dep in all_deps:
        node_info = _get_node_info(dep)
        node_data.append(node_info)

    # Generate edge data
    edge_data = [
        {
            "from": f"n{id(node.parent)}",
            "to": f"n{id(node.node)}",
            "from_block": node.parent.block_name,
            "to_block": node.node.block_name,
        }
        for node in nodes
        if node.parent is not None
    ]

    # Get block information
    blocks = list({node.node.block_name for node in nodes if node.node.block_name is not None})

    # Generate legend
    legend = ASGI_ENGIN_LEGEND if isinstance(instance, ASGIEngin) else DEFAULT_LEGEND

    return {
        "nodes": node_data,
        "edges": edge_data,
        "blocks": blocks,
        "legend": legend,
        "app_origin": _APP_ORIGIN,
    }


def _get_node_info(node: Dependency) -> dict[str, Any]:
    """Extract node information for JSON representation."""
    node_id = f"n{id(node)}"  # Add 'n' prefix to match mermaid node IDs
    label = ""
    style_classes = []

    # Determine if external
    node_root_package = node.source_package.split(".", maxsplit=1)[0]
    is_external = node_root_package != _APP_ORIGIN
    if is_external:
        style_classes.append("external")

    # Collect detailed information for tooltips
    details: dict[str, Any] = {
        "full_name": node.name,
        "source_module": node.source_module,
        "source_package": node.source_package,
        "parameters": [],
        "return_type": None,
        "scope": None,
    }

    # Get parameter information
    if hasattr(node, "parameter_type_ids"):
        details["parameters"] = [str(param_id) for param_id in node.parameter_type_ids]

    # Determine node type and extract specific details
    if isinstance(node, Supply):
        node_type = "Supply"
        label += f"{_short_name(node.return_type_id)}"
        shape = "round"
        details["return_type"] = str(node.return_type_id)
        if hasattr(node, "_value"):
            details["value_type"] = type(node._value).__name__
    elif isinstance(node, Provide):
        node_type = "Provide"
        label += f"{_short_name(node.return_type_id)}"
        shape = "rect"
        details["return_type"] = str(node.return_type_id)
        details["factory_function"] = node.func_name
        if node.scope:
            details["scope"] = node.scope
            style_classes.append(f"scope-{node.scope}")
        if node.is_multiprovider:
            details["multiprovider"] = True
            style_classes.append("multi")
    elif isinstance(node, Entrypoint):
        node_type = "Entrypoint"
        entrypoint_type = node.parameter_type_ids[0]
        label += f"{entrypoint_type}"
        shape = "trapezoid"
        details["entrypoint_type"] = str(entrypoint_type)
    elif isinstance(node, Invoke):
        node_type = "Invoke"
        label += f"{node.func_name}"
        shape = "trapezoid"
        details["function"] = node.func_name
    elif APIRouteDependency is not None and isinstance(node, APIRouteDependency):
        node_type = "APIRoute"
        label += f"{node.name}"
        shape = "subroutine"
        if hasattr(node, "route"):
            details["methods"] = (
                list(node.route.methods) if hasattr(node.route, "methods") else []
            )
            details["path"] = getattr(node.route, "path", "")
    else:
        node_type = "Other"
        label += f"{node.name}"
        shape = "rect"

    return {
        "id": node_id,
        "label": label,
        "type": node_type,
        "external": is_external,
        "block": node.block_name,
        "shape": shape,
        "style_classes": style_classes,
        "source_module": node.source_module,
        "source_package": node.source_package,
        "details": details,
    }


def _short_name(name: TypeId) -> str:
    return str(name).rsplit(".", maxsplit=1)[-1]


_GRAPH_HTML = (Path(__file__).parent / "_graph.html").read_text()

DEFAULT_LEGEND = (
    "0[/Invoke/] ~~~ 1[/Entrypoint\\] ~~~ 2[Provide] ~~~ 3(Supply)"
    ' ~~~ 4["`Block Grouping`"]:::b0 ~~~ 5[External Dependency]:::external'
)
ASGI_ENGIN_LEGEND = DEFAULT_LEGEND + " ~~~ 6[[API Route]]"



from collections.abc import Iterable
from dataclasses import dataclass

from engin import Provide
from engin._dependency import Dependency
from engin._type_utils import TypeId


@dataclass(slots=True, frozen=True, kw_only=True)
class Node:
    """
    A Node in the Dependency Graph.
    """

    node: Dependency
    parent: Dependency | None

    def __repr__(self) -> str:
        return f"Node(node={self.node!s},parent={self.parent!s})"


class DependencyGrapher:
    def __init__(self, providers: dict[TypeId, Provide | list[Provide]]) -> None:
        self._providers: dict[TypeId, Provide | list[Provide]] = providers

    def resolve(self, roots: Iterable[Dependency]) -> list[Node]:
        return self._resolve_recursive(roots, seen=set())

    def _resolve_recursive(
        self, roots: Iterable[Dependency], *, seen: set[TypeId]
    ) -> list[Node]:
        nodes: list[Node] = []
        for root in roots:
            for parameter in root.parameter_type_ids:
                provider = self._providers[parameter]

                # multiprovider
                if isinstance(provider, list):
                    nodes.extend(Node(node=p, parent=root) for p in provider)
                    if parameter not in seen:
                        nodes.extend(self._resolve_recursive(provider, seen=seen))
                # single provider
                else:
                    nodes.append(Node(node=provider, parent=root))
                    if parameter not in seen:
                        nodes.extend(self._resolve_recursive([provider], seen=seen))

                seen.add(parameter)

        return nodes



import inspect
import typing
from abc import ABC
from collections.abc import Awaitable, Callable
from inspect import Parameter, Signature, isclass, iscoroutinefunction
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    ParamSpec,
    TypeAlias,
    TypeVar,
    cast,
    get_type_hints,
)

from engin._introspect import get_first_external_frame
from engin._option import Option
from engin._type_utils import TypeId

if TYPE_CHECKING:
    from engin._engin import Engin

P = ParamSpec("P")
T = TypeVar("T")
Func: TypeAlias = Callable[P, T]


def _noop(*args: Any, **kwargs: Any) -> None: ...


class Dependency(ABC, Option, Generic[P, T]):
    def __init__(self, func: Func[P, T]) -> None:
        self._func = func
        self._is_async = iscoroutinefunction(func)
        self._signature = inspect.signature(self._func, eval_str=True)
        self._block_name: str | None = None

        source_frame = get_first_external_frame()
        self._source_package = cast("str", source_frame.frame.f_globals["__package__"])
        self._source_frame = cast("str", source_frame.frame.f_globals["__name__"])

    @property
    def source_module(self) -> str:
        """
        The module that this Dependency originated from.

        Returns:
            A string, e.g. "examples.fastapi.app"
        """
        return self._source_frame

    @property
    def source_package(self) -> str:
        """
        The package that this Dependency originated from.

        Returns:
            A string, e.g. "engin"
        """
        return self._source_package

    @property
    def block_name(self) -> str | None:
        return self._block_name

    @property
    def func_name(self) -> str:
        return self._func.__name__

    @property
    def name(self) -> str:
        if self._block_name:
            return f"{self._block_name}.{self._func.__name__}"
        else:
            return f"{self._func.__module__}.{self._func.__name__}"

    @property
    def parameter_type_ids(self) -> list[TypeId]:
        parameters = list(self._signature.parameters.values())
        if not parameters:
            return []
        if parameters[0].name == "self":
            parameters.pop(0)
        return [TypeId.from_type(param.annotation) for param in parameters]

    @property
    def signature(self) -> Signature:
        return self._signature

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if self._is_async:
            return await cast("Awaitable[T]", self._func(*args, **kwargs))
        else:
            return self._func(*args, **kwargs)


class Invoke(Dependency):
    """
    Marks a function as an Invocation.

    Invocations are functions that are called prior to lifecycle startup. Invocations
    should not be long running as the application startup will be blocked until all
    Invocation are completed.

    Invocations can be provided as an Option to the Engin or a Block.

    Examples:
        ```python3
        def print_string(a_string: str) -> None:
            print(f"invoking with value: '{a_string}'")

        invocation = Invoke(print_string)
        ```
    """

    def __init__(self, invocation: Func[P, T]) -> None:
        super().__init__(func=invocation)

    def apply(self, engin: "Engin") -> None:
        engin._invocations.append(self)

    def __str__(self) -> str:
        return f"Invoke({self.name})"


class Entrypoint(Invoke):
    """
    Marks a type as an Entrypoint.

    Entrypoints are a short hand for no-op Invocations that can be used to
    """

    def __init__(self, type_: type[Any]) -> None:
        self._type = type_
        super().__init__(invocation=_noop)

    @property
    def parameter_type_ids(self) -> list[TypeId]:
        return [TypeId.from_type(self._type)]

    @property
    def signature(self) -> Signature:
        return Signature(
            parameters=[
                Parameter(name="x", kind=Parameter.POSITIONAL_ONLY, annotation=self._type)
            ]
        )

    def __str__(self) -> str:
        return f"Entrypoint({TypeId.from_type(self._type)})"


class Provide(Dependency[Any, T]):
    def __init__(
        self,
        factory: Func[P, T],
        *,
        scope: str | None = None,
        as_type: type | None = None,
        override: bool = False,
    ) -> None:
        """
        Provide a type via a factory function.

        Args:
            factory: the factory function that returns the type.
            scope: (optional) associate this provider with a specific scope.
            as_type: (optional) allows you to explicitly specify the provided type, e.g.
                to type erase a concrete type, or to provide a mock implementation.
            override: (optional) allow this provider to override other providers for the
                same type from the same package.
        """
        if not callable(factory):
            msg = "Provided value is not callable, did you mean to use Supply instead?"
            raise ValueError(msg)
        super().__init__(func=factory)
        self._scope = scope
        self._override = override
        self._explicit_type = as_type
        self._return_type = self._resolve_return_type()
        self._return_type_id = TypeId.from_type(self._return_type)

        if self._explicit_type is not None:
            self._signature = self._signature.replace(return_annotation=self._explicit_type)

        self._is_multi = (
            typing.get_origin(self._return_type) is list or self._return_type is list
        )

        # Validate that the provider does to depend on its own output value, as this will
        # cause a recursion error and is undefined behaviour wise.
        if any(
            self._return_type == param.annotation
            for param in self._signature.parameters.values()
        ):
            raise ValueError("A provider cannot depend on its own return type")

        # Validate that multiproviders only return a list of one type.
        if self._is_multi:
            args = typing.get_args(self._return_type)
            if len(args) != 1:
                msg = (
                    "A multiprovider must be of the form list[X], not "
                    f"'{self._return_type_id}'"
                )
                raise ValueError(msg)

    @property
    def return_type(self) -> type[T]:
        return self._return_type

    @property
    def return_type_id(self) -> TypeId:
        return self._return_type_id

    @property
    def is_multiprovider(self) -> bool:
        return self._is_multi

    @property
    def scope(self) -> str | None:
        return self._scope

    def apply(self, engin: "Engin") -> None:
        type_id = self.return_type_id
        if self.is_multiprovider:
            engin._multiproviders[type_id].append(self)
            return

        if type_id not in engin._providers:
            engin._providers[type_id] = self
            return

        existing_provider = engin._providers[type_id]
        is_same_package = existing_provider.source_package == self.source_package

        # overwriting a dependency from the same package must be explicit
        if is_same_package and not self._override:
            msg = (
                f"{self} from '{self._source_frame}' is implicitly overriding "
                f"{existing_provider} from '{existing_provider.source_module}', if this "
                "is intentional specify `override=True` for the overriding Provider"
            )
            raise RuntimeError(msg)

        engin._providers[type_id] = self

    def __hash__(self) -> int:
        return hash(self.return_type_id)

    def __str__(self) -> str:
        return f"Provide(factory={self.func_name}, type={self._return_type_id})"

    def _resolve_return_type(self) -> type[T]:
        if self._explicit_type is not None:
            return self._explicit_type
        if isclass(self._func):
            return_type = self._func  # __init__ returns self
        else:
            try:
                return_type = get_type_hints(self._func, include_extras=True)["return"]
            except KeyError as err:
                raise RuntimeError(
                    f"Dependency '{self.name}' requires a return typehint"
                ) from err

        return return_type


class Supply(Provide, Generic[T]):
    def __init__(
        self, value: T, *, as_type: type | None = None, override: bool = False
    ) -> None:
        """
        Supply a value.

        This is a shorthand which under the hood creates a Provider with a noop factory
        function.

        Args:
            value: the value to Supply.
            as_type: (optional) allows you to explicitly specify the provided type, e.g.
              to type erase a concrete type, or to provide a mock implementation.
            override: (optional) allow this provider to override other providers for the
              same type from the same package.
        """
        self._value = value
        super().__init__(factory=self._get_val, as_type=as_type, override=override)

    @property
    def name(self) -> str:
        if self._block_name:
            return f"{self._block_name}.supply"
        else:
            return f"{self._source_frame}.supply"

    def _resolve_return_type(self) -> type[T]:
        if self._explicit_type is not None:
            return self._explicit_type
        if isinstance(self._value, list):
            return list[type(self._value[0])]  # type: ignore[misc,return-value]
        return type(self._value)

    def _get_val(self) -> T:
        return self._value

    def __str__(self) -> str:
        return f"Supply(value={self._value}, type={self.return_type_id})"



import contextlib
import traceback
from collections.abc import AsyncIterator, Awaitable, Callable, MutableMapping
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from typing import Any, ClassVar, Protocol, TypeAlias

from engin import Engin, Entrypoint, Option, Supply

__all__ = ["ASGIEngin", "ASGIType", "engin_to_lifespan"]

from engin._graph import DependencyGrapher, Node

_Scope: TypeAlias = MutableMapping[str, Any]
_Message: TypeAlias = MutableMapping[str, Any]
_Receive: TypeAlias = Callable[[], Awaitable[_Message]]
_Send: TypeAlias = Callable[[_Message], Awaitable[None]]


class ASGIType(Protocol):
    async def __call__(self, scope: _Scope, receive: _Receive, send: _Send) -> None: ...


class ASGIEngin(Engin, ASGIType):
    _STOP_ON_SINGAL = False  # web server implementation is responsible for this

    _asgi_type: ClassVar[type[ASGIType]] = ASGIType  # type: ignore[type-abstract]
    _asgi_app: ASGIType

    def __init__(self, *options: Option) -> None:
        super().__init__(*options)

        if not self._assembler.has(self._asgi_type):
            raise LookupError(
                f"A provider for `{self._asgi_type.__name__}` was expected, none found"
            )

    async def __call__(self, scope: _Scope, receive: _Receive, send: _Send) -> None:
        if scope["type"] == "lifespan":
            message = await receive()
            receive = _Rereceive(message)
            if message["type"] == "lifespan.startup":
                try:
                    await self._startup()
                except Exception as err:
                    exc = "".join(traceback.format_exception(err))
                    await send({"type": "lifespan.startup.failed", "message": exc})
                    raise

            elif message["type"] == "lifespan.shutdown":
                await self.stop()

        with self._assembler.scope("request"):
            await self._asgi_app(scope, receive, send)

    async def _startup(self) -> None:
        self._asgi_app = await self._assembler.build(self._asgi_type)
        await self.start()

    def graph(self) -> list[Node]:
        grapher = DependencyGrapher({**self._providers, **self._multiproviders})
        return grapher.resolve([Entrypoint(self._asgi_type), *self._invocations])


class _Rereceive:
    def __init__(self, message: _Message) -> None:
        self._message = message

    async def __call__(self, *args: Any, **kwargs: Any) -> _Message:
        return self._message


def engin_to_lifespan(engin: Engin) -> Callable[[ASGIType], AbstractAsyncContextManager[None]]:
    """
    Transforms the Engin instance into an ASGI lifespan task.

    This is to enable users to use the Engin framework with existing ASGI applications,
    where it is not desired to replace the ASGI application with an ASGIEngin.

    Args:
        engin: the engin instance to transform.

    Returns:
        An ASGI lifespan task.
    """

    @asynccontextmanager
    async def engin_lifespan(app: ASGIType) -> AsyncIterator[None]:
        # ensure the Engin
        with contextlib.suppress(ValueError):
            engin.assembler.add(Supply(app))

        app.state.assembler = engin.assembler  # type: ignore[attr-defined]

        await engin.start()
        yield
        await engin.stop()

    return engin_lifespan



import asyncio
import logging
from collections import defaultdict
from collections.abc import Iterable, Sequence
from contextvars import ContextVar
from dataclasses import dataclass
from inspect import BoundArguments, Signature
from types import TracebackType
from typing import Any, Generic, TypeVar, cast

from typing_extensions import Self

from engin._dependency import Dependency, Provide, Supply
from engin._type_utils import TypeId
from engin.exceptions import NotInScopeError, ProviderError, TypeNotProvidedError

LOG = logging.getLogger("engin")

T = TypeVar("T")
_SCOPE: ContextVar[list[str] | None] = ContextVar("_SCOPE", default=None)


def _get_scope() -> list[str]:
    if _SCOPE.get() is None:
        _SCOPE.set([])
    return cast("list[str]", _SCOPE.get())


@dataclass(slots=True, kw_only=True, frozen=True)
class AssembledDependency(Generic[T]):
    """
    An AssembledDependency can be called to construct the result.
    """

    dependency: Dependency[Any, T]
    bound_args: BoundArguments

    async def __call__(self) -> T:
        """
        Construct the dependency.

        Returns:
            The constructed value.
        """
        return await self.dependency(*self.bound_args.args, **self.bound_args.kwargs)


class Assembler:
    """
    A container for Providers that is responsible for building provided types.

    The Assembler acts as a cache for previously built types, meaning repeat calls
    to `build` will produce the same value.

    Examples:
        ```python
        def build_str() -> str:
            return "foo"

        a = Assembler([Provide(build_str)])
        await a.build(str)
        ```
    """

    def __init__(self, providers: Iterable[Provide]) -> None:
        self._providers: dict[TypeId, Provide[Any]] = {}
        self._multiproviders: dict[TypeId, list[Provide[list[Any]]]] = defaultdict(list)
        self._assembled_outputs: dict[TypeId, Any] = {}
        self._lock = asyncio.Lock()
        self._graph_cache: dict[TypeId, list[Provide]] = defaultdict(list)

        for provider in providers:
            type_id = provider.return_type_id
            if not provider.is_multiprovider:
                if type_id in self._providers:
                    raise RuntimeError(f"A Provider already exists for '{type_id}'")
                self._providers[type_id] = provider
            else:
                self._multiproviders[type_id].append(provider)

    @classmethod
    def from_mapped_providers(
        cls,
        providers: dict[TypeId, Provide[Any]],
        multiproviders: dict[TypeId, list[Provide[list[Any]]]],
    ) -> Self:
        """
        Create an Assembler from pre-mapped providers.

        This method is only exposed for performance reasons in the case that Providers
        have already been mapped, it is recommended to use the `__init__` method if this
        is no the case.

        Args:
            providers: a dictionary of Providers with the Provider's `return_type_id` as
              the key.
            multiproviders: a dictionary of list of Providers with the Provider's
              `return_type_id` as key. All Providers in the given list must be for the
              related `return_type_id`.

        Returns:
            An Assembler instance.
        """
        assembler = cls(tuple())  # noqa: C408
        assembler._providers = providers
        assembler._multiproviders = multiproviders
        return assembler

    @property
    def providers(self) -> Sequence[Provide[Any]]:
        multi_providers = [p for multi in self._multiproviders.values() for p in multi]
        return [*self._providers.values(), *multi_providers]

    async def assemble(self, dependency: Dependency[Any, T]) -> AssembledDependency[T]:
        """
        Assemble a dependency.

        Given a Dependency type, such as Invoke, the Assembler constructs the types
        required by the Dependency's signature from its providers.

        Args:
            dependency: the Dependency to assemble.

        Returns:
            An AssembledDependency, which can be awaited to construct the final value.
        """
        async with self._lock:
            return AssembledDependency(
                dependency=dependency,
                bound_args=await self._bind_arguments(dependency.signature),
            )

    async def build(self, type_: type[T]) -> T:
        """
        Build the type from Assembler's factories.

        If the type has been built previously the value will be cached and will return the
        same instance.

        Args:
            type_: the type of the desired value to build.

        Raises:
            TypeNotProvidedError: When no provider is found for the given type.
            ProviderError: When a provider errors when trying to construct the type or
                any of its dependent types.

        Returns:
            The constructed value.
        """
        type_id = TypeId.from_type(type_)
        if type_id in self._assembled_outputs:
            return cast("T", self._assembled_outputs[type_id])
        if type_id.multi:
            if type_id not in self._multiproviders:
                raise TypeNotProvidedError(type_id)

            out = []
            for provider in self._multiproviders[type_id]:
                if provider.scope and provider.scope not in _get_scope():
                    raise NotInScopeError(provider=provider, scope_stack=_get_scope())
                assembled_dependency = await self.assemble(provider)
                try:
                    out.extend(await assembled_dependency())
                except Exception as err:
                    raise ProviderError(
                        provider=provider,
                        error_type=type(err),
                        error_message=str(err),
                    ) from err
            self._assembled_outputs[type_id] = out
            return out  # type: ignore[return-value]
        else:
            if type_id not in self._providers:
                raise TypeNotProvidedError(type_id)

            provider = self._providers[type_id]
            if provider.scope and provider.scope not in _get_scope():
                raise NotInScopeError(provider=provider, scope_stack=_get_scope())

            assembled_dependency = await self.assemble(provider)
            try:
                value = await assembled_dependency()
            except Exception as err:
                raise ProviderError(
                    provider=provider,
                    error_type=type(err),
                    error_message=str(err),
                ) from err
            self._assembled_outputs[type_id] = value
            return value  # type: ignore[return-value]

    def has(self, type_: type[T]) -> bool:
        """
        Returns True if this Assembler has a provider for the given type.

        Args:
            type_: the type to check.

        Returns:
            True if the Assembler has a provider for type else False.
        """
        type_id = TypeId.from_type(type_)
        if type_id.multi:
            return type_id in self._multiproviders
        else:
            return type_id in self._providers

    def add(self, provider: Provide) -> None:
        """
        Add a provider to the Assembler post-initialisation.

        If this replaces an existing provider, this will clear all previously assembled
        output. Note: multiproviders cannot be replaced, they are always appended.

        Args:
            provider: the Provide instance to add.

        Returns:
             None
        """
        type_id = provider.return_type_id
        if provider.is_multiprovider:
            self._multiproviders[type_id].append(provider)
        else:
            self._providers[type_id] = provider

        self._assembled_outputs.clear()
        self._graph_cache.clear()

    def scope(self, scope: str) -> "_ScopeContextManager":
        return _ScopeContextManager(scope=scope, assembler=self)

    def _exit_scope(self, scope: str) -> None:
        for type_id, provider in self._providers.items():
            if provider.scope == scope:
                self._assembled_outputs.pop(type_id, None)

    def _resolve_providers(self, type_id: TypeId, resolved: set[TypeId]) -> Iterable[Provide]:
        """
        Resolves the chain of providers required to satisfy the provider of a given type.
        Ordering of the return value is very important here!
        """
        if type_id in self._graph_cache:
            return self._graph_cache[type_id]

        if type_id.multi:
            root_providers = self._multiproviders.get(type_id)
        else:
            root_providers = [provider] if (provider := self._providers.get(type_id)) else None

        if not root_providers:
            if type_id.multi:
                LOG.warning(f"no provider for '{type_id}' defaulting to empty list")
                root_providers = [(Supply([], as_type=list[type_id.type]))]  # type: ignore[name-defined]
                # store default to prevent the warning appearing multiple times
                self._multiproviders[type_id] = root_providers
            else:
                raise TypeNotProvidedError(type_id)

        # providers that must be satisfied to satisfy the root level providers
        resolved_providers = [
            child_provider
            for root_provider in root_providers
            for root_provider_param in root_provider.parameter_type_ids
            for child_provider in self._resolve_providers(root_provider_param, resolved)
            if root_provider_param not in resolved
        ]

        resolved_providers.extend(root_providers)

        resolved.add(type_id)
        self._graph_cache[type_id] = resolved_providers

        return resolved_providers

    async def _satisfy(self, target: TypeId) -> None:
        for provider in self._resolve_providers(target, set()):
            if (
                not provider.is_multiprovider
                and provider.return_type_id in self._assembled_outputs
            ):
                continue
            type_id = provider.return_type_id

            bound_args = await self._bind_arguments(provider.signature)
            try:
                value = await provider(*bound_args.args, **bound_args.kwargs)
            except Exception as err:
                raise ProviderError(
                    provider=provider, error_type=type(err), error_message=str(err)
                ) from err

            if provider.is_multiprovider:
                if type_id in self._assembled_outputs:
                    self._assembled_outputs[type_id].extend(value)
                else:
                    self._assembled_outputs[type_id] = value
            else:
                self._assembled_outputs[type_id] = value

    async def _bind_arguments(self, signature: Signature) -> BoundArguments:
        args = []
        kwargs = {}
        for param_name, param in signature.parameters.items():
            if param_name == "self":
                args.append(object())
                continue
            param_key = TypeId.from_type(param.annotation)
            if param_key not in self._assembled_outputs:
                await self._satisfy(param_key)
            val = self._assembled_outputs[param_key]
            if param.kind == param.POSITIONAL_ONLY:
                args.append(val)
            else:
                kwargs[param.name] = val

        return signature.bind(*args, **kwargs)


class _ScopeContextManager:
    def __init__(self, scope: str, assembler: Assembler) -> None:
        self._scope = scope
        self._assembler = assembler

    def __enter__(self) -> Assembler:
        _get_scope().append(self._scope)
        return self._assembler

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None:
        popped = _get_scope().pop()
        if popped != self._scope:
            raise RuntimeError(
                f"Exited scope '{popped}' is not the expected scope '{self._scope}'"
            )
        self._assembler._exit_scope(self._scope)



import inspect
from collections.abc import Callable, Iterable, Sequence
from itertools import chain
from typing import TYPE_CHECKING, ClassVar

from engin._dependency import Dependency, Func, Invoke, Provide
from engin._option import Option
from engin.exceptions import InvalidBlockError

if TYPE_CHECKING:
    from engin._engin import Engin

_BUILTIN_CLASS_FUNCTIONS = ("__annotate_func__",)


def provide(
    func_: Func | None = None, *, scope: str | None = None, override: bool = False
) -> Func | Callable[[Func], Func]:
    """
    A decorator for defining a Provider in a Block.
    """

    def _inner(func: Func) -> Func:
        func._opt = Provide(func, override=override, scope=scope)  # type: ignore[attr-defined]
        return func

    if func_ is None:
        return _inner
    else:
        return _inner(func_)


def invoke(func_: Func | None = None) -> Func | Callable[[Func], Func]:
    """
    A decorator for defining an Invocation in a Block.
    """

    def _inner(func: Func) -> Func:
        func._opt = Invoke(func)  # type: ignore[attr-defined]
        return func

    if func_ is None:
        return _inner
    else:
        return _inner(func_)


class Block:
    """
    A Block is a collection of providers and invocations.

    Blocks are useful for grouping a collection of related providers and invocations, and
    are themselves a valid Option type that can be passed to the Engin.

    Providers are defined as methods decorated with the `provide` decorator, and similarly
    for Invocations and the `invoke` decorator.

    Examples:
        Define a simple block.

        ```python3
        from engin import Block, provide, invoke

        class MyBlock(Block):
            @provide
            def some_str(self) -> str:
                return "foo"

            @invoke
            def print_str(self, string: str) -> None:
                print(f"invoked on string '{string}')
        ```
    """

    name: ClassVar[str | None] = None
    options: ClassVar[Sequence[Option]] = []

    @classmethod
    def apply(cls, engin: "Engin") -> None:
        block_name = cls.name or cls.__name__
        for option in chain(cls.options, cls._method_options()):
            if isinstance(option, Dependency):
                option._block_name = block_name
            option.apply(engin)

    @classmethod
    def _method_options(cls) -> Iterable[Provide | Invoke]:
        for name, method in inspect.getmembers(cls, inspect.isfunction):
            if name in _BUILTIN_CLASS_FUNCTIONS:
                continue
            if option := getattr(method, "_opt", None):
                if not isinstance(option, Provide | Invoke):
                    raise InvalidBlockError(
                        block=cls,
                        reason="Block option is not an instance of Provide or Invoke",
                    )
                yield option
            else:
                raise InvalidBlockError(
                    block=cls,
                    reason=(
                        f"Method '{name}' is not a Provider or Invocation, did you "
                        "forget to decorate it?"
                    ),
                )



from typing import Annotated

import typer
from rich.console import Console

from engin._cli._common import COMMON_HELP, get_engin_instance
from engin.exceptions import TypeNotProvidedError

cli = typer.Typer()


@cli.command(name="check")
def check_dependencies(
    app: Annotated[
        str | None,
        typer.Argument(help=COMMON_HELP["app"]),
    ] = None,
) -> None:
    """
    Validates that all dependencies are satisfied for the given engin instance.

    This command checks that all providers required by invocations and other providers
    are available. It's intended for use in CI to catch missing dependencies.

    Examples:

        1. `engin check`

    Returns:
        Exit code 0 if all dependencies are satisfied.
        Exit code 1 if there are missing providers.
    """
    _, _, instance = get_engin_instance(app)

    console = Console()
    assembler = instance.assembler
    missing_providers = set()

    for invocation in instance._invocations:
        for param_type_id in invocation.parameter_type_ids:
            try:
                assembler._resolve_providers(param_type_id, set())
            except TypeNotProvidedError:
                missing_providers.add(param_type_id)

    if missing_providers:
        sorted_missing = sorted(str(type_id) for type_id in missing_providers)

        console.print("❌ Missing providers found:", style="red bold")
        for missing_type in sorted_missing:
            console.print(f"  • {missing_type}", style="red")

        raise typer.Exit(code=1)
    else:
        console.print("✅ All dependencies are satisfied!", style="green bold")
        raise typer.Exit(code=0)



from typing import TYPE_CHECKING, Any

from engin._dependency import Provide
from engin._type_utils import TypeId

if TYPE_CHECKING:
    from engin._block import Block


class EnginError(Exception):
    """
    Base class for all custom exceptions in the Engin library.
    """


class InvalidBlockError(EnginError):
    """
    Raised when an invalid block is instantiated.
    """

    def __init__(self, block: "type[Block]", reason: str) -> None:
        self.block = block
        self.block_name = block.name or block.__name__
        self.message = f"block '{self.block_name}' is invalid, reason: '{reason}'"

    def __str__(self) -> str:
        return self.message


class AssemblerError(EnginError):
    """
    Base class for all custom exceptions raised by the Assembler.
    """


class TypeNotProvidedError(AssemblerError):
    """
    Raised when the Assembler cannot assemble a type due to a missing Provider.
    """

    def __init__(self, type_id: TypeId) -> None:
        self.type_id = type_id
        self.message = f"no provider found for '{type_id}'"

    def __str__(self) -> str:
        return self.message


class ProviderError(AssemblerError):
    """
    Raised when a Provider errors during Assembly.
    """

    def __init__(
        self,
        provider: Provide[Any],
        error_type: type[Exception],
        error_message: str,
    ) -> None:
        self.provider = provider
        self.error_type = error_type
        self.error_message = error_message
        self.message = (
            f"provider '{provider.name}' errored with error "
            f"({error_type.__name__}): '{error_message}'"
        )

    def __str__(self) -> str:
        return self.message


class NotInScopeError(AssemblerError):
    """
    Raised when a Provider is requested outside of its scope.
    """

    def __init__(self, provider: Provide[Any], scope_stack: list[str]) -> None:
        self.provider = provider
        self.message = (
            f"provider '{provider.name}' was requested outside of its specified scope "
            f"'{provider.scope}', current scope stack is {scope_stack}"
        )

    def __str__(self) -> str:
        return self.message


__all__ = [
    "AssemblerError",
    "EnginError",
    "InvalidBlockError",
    "NotInScopeError",
    "ProviderError",
    "TypeNotProvidedError",
]



import asyncio
import logging
from collections.abc import Awaitable, Callable
from contextlib import AbstractAsyncContextManager, AbstractContextManager
from inspect import iscoroutinefunction
from types import TracebackType
from typing import TypeAlias, TypeGuard, cast

LOG = logging.getLogger("engin")

_AnyContextManager: TypeAlias = AbstractAsyncContextManager | AbstractContextManager
_ParameterlessCallable: TypeAlias = Callable[[], None | Awaitable[None]]


class Lifecycle:
    """
    Allows dependencies to define startup and shutdown tasks for the application.

    Lifecycle tasks are defined using Context Managers, these can be async or sync.

    Lifecycle tasks should generally be defined in Providers as they are tied to the
    construction of a given dependency, but can be used in Invocations. The Lifecycle
    type is provided as a built-in Dependency by the Engin framework.

    Examples:
        Using a type that implements context management.

        ```python
        from httpx import AsyncClient

        def my_provider(lifecycle: Lifecycle) -> AsyncClient:
            client = AsyncClient()

            # AsyncClient is a context manager
            lifecycle.append(client)
        ```

        Defining a custom lifecycle.

        ```python
        def my_provider(lifecycle: Lifecycle) -> str:
            @contextmanager
            def task():
                print("starting up!")
                yield
                print("shutting down!)

            lifecycle.append(task)
        ```

        Defining a custom lifecycle using a LifecycleHook.

        ```python
        def my_provider(lifecycle: Lifecycle) -> str:
            connection_pool = ConnectionPool()

            lifecycle.hook(
                on_start=connection_pool.connect,
                on_stop=connection_pool.close,
            )
        ```
    """

    def __init__(self) -> None:
        self._context_managers: list[AbstractAsyncContextManager] = []

    def append(self, cm: _AnyContextManager, /) -> None:
        """
        Append a Lifecycle task to the list.

        Args:
            cm: a task defined as a ContextManager or AsyncContextManager.
        """
        suppressed_cm = _AExitSuppressingAsyncContextManager(cm)
        self._context_managers.append(suppressed_cm)

    def hook(
        self,
        *,
        on_start: _ParameterlessCallable | None = None,
        on_stop: _ParameterlessCallable | None = None,
    ) -> None:
        """
        Append a hook to the Lifecycle.

        At least one of `on_start` or `on_stop` must be provided.

        Args:
            on_start: a callable to be executed on Lifecycle startup.
            on_stop: a callable to be executed on Lifecycle shutdown.
        """
        if on_start is None and on_stop is None:
            raise ValueError("At least one of on_start or on_stop must be provided")
        self.append(LifecycleHook(on_start=on_start, on_stop=on_stop))

    def list(self) -> list[AbstractAsyncContextManager]:
        """
        List all the defined tasks.

        Returns:
            A copy of the list of Lifecycle tasks.
        """
        return self._context_managers[:]


class LifecycleHook(AbstractAsyncContextManager):
    def __init__(
        self,
        on_start: _ParameterlessCallable | None = None,
        on_stop: _ParameterlessCallable | None = None,
    ) -> None:
        self._on_start = on_start
        self._on_stop = on_stop

    async def __aenter__(self) -> None:
        if self._on_start is not None:
            func = self._on_start
            if iscoroutinefunction(func):
                await func()
            else:
                await asyncio.to_thread(func)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None:
        if self._on_stop is not None:
            func = self._on_stop
            if iscoroutinefunction(func):
                await func()
            else:
                await asyncio.to_thread(func)


class _AExitSuppressingAsyncContextManager(AbstractAsyncContextManager):
    def __init__(self, cm: _AnyContextManager) -> None:
        self._cm = cm

    async def __aenter__(self) -> None:
        if self._is_async_cm(self._cm):
            await self._cm.__aenter__()
        else:
            await asyncio.to_thread(cast("AbstractContextManager", self._cm).__enter__)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
        /,
    ) -> None:
        try:
            if self._is_async_cm(self._cm):
                await self._cm.__aexit__(exc_type, exc_value, traceback)
            else:
                await asyncio.to_thread(
                    cast("AbstractContextManager", self._cm).__exit__,
                    exc_type,
                    exc_value,
                    traceback,
                )
        except Exception as err:
            LOG.error("error in lifecycle hook stop, ignoring...", exc_info=err)

    @staticmethod
    def _is_async_cm(cm: _AnyContextManager) -> TypeGuard[AbstractAsyncContextManager]:
        return hasattr(cm, "__aenter__")



import inspect
from collections.abc import Iterable
from inspect import FrameInfo


def walk_stack() -> Iterable[FrameInfo]:
    """
    Fast alternative to `inspect.stack()`

    Compared to `inspect.stack()`:
     - Does not read source files to load neighboring context
     - Less accurate filename determination, still correct for most cases
     - Does not compute 3.11+ code positions (PEP 657)
    """

    frame = inspect.currentframe()

    while frame := frame and frame.f_back:
        yield inspect.FrameInfo(
            frame,
            inspect.getfile(frame),
            frame.f_lineno,
            frame.f_code.co_name,
            None,
            None,
        )


def get_first_external_frame() -> FrameInfo:
    for frame_info in walk_stack():
        frame = frame_info.frame
        if frame.f_globals["__package__"] != "engin" or frame.f_back is None:
            return frame_info
    raise RuntimeError("Unable to find external frame")



import asyncio
import logging
import signal
import sys
from asyncio import Event
from collections import defaultdict
from contextlib import AsyncExitStack
from enum import Enum
from itertools import chain
from types import FrameType
from typing import TYPE_CHECKING, ClassVar

from anyio import create_task_group, open_signal_receiver

from engin._assembler import AssembledDependency, Assembler
from engin._dependency import Invoke, Provide, Supply
from engin._graph import DependencyGrapher, Node
from engin._lifecycle import Lifecycle
from engin._option import Option
from engin._supervisor import Supervisor
from engin.exceptions import EnginError

if TYPE_CHECKING:
    from engin._type_utils import TypeId

_OS_IS_WINDOWS = sys.platform == "win32"
LOG = logging.getLogger("engin")


class _EnginState(Enum):
    IDLE = 0
    """
    Engin is not yet started.
    """

    RUNNING = 1
    """
    Engin is currently running.
    """

    SHUTDOWN = 2
    """
    Engin has performed shutdown
    """


class Engin:
    """
    The Engin is a modular application defined by a collection of options.

    Users should instantiate the Engin with a number of options, where options can be an
    instance of Provide, Invoke, or a collection of these combined in a Block.

    To create a useful application, users should pass in one or more providers (Provide or
    Supply) and at least one invocation (Invoke or Entrypoint).

    When instantiated the Engin can be run. This is typically done via the `run` method,
    but for advanced usecases it can be easier to use the `start` and `stop` methods.

    When ran the Engin takes care of the complete application lifecycle:

    1. The Engin assembles all Invocations. Only Providers that are required to satisfy
       the Invoke options parameters are assembled.
    2. All Invocations are run sequentially in the order they were passed in to the Engin.
    3. Lifecycle Startup tasks registered by assembled dependencies are run sequentially.
    4. The Engin waits for a stop signal, i.e. SIGINT or SIGTERM, or a supervised task that
       causes a shutdown.
    5. Lifecyce Shutdown tasks are run in the reverse order to the Startup order.

    Examples:
        ```python
        import asyncio

        from httpx import AsyncClient

        from engin import Engin, Invoke, Lifecycle, Provide


        def httpx_client(lifecycle: Lifecycle) -> AsyncClient:
            client = AsyncClient()
            lifecycle.append(client)
            return client


        async def main(http_client: AsyncClient) -> None:
            print(await http_client.get("https://httpbin.org/get"))

        engin = Engin(Provide(httpx_client), Invoke(main))

        asyncio.run(engin.run())
        ```
    """

    _LIB_OPTIONS: ClassVar[list[Option]] = [Provide(Lifecycle), Provide(Supervisor)]
    _STOP_ON_SINGAL: ClassVar[bool] = True

    def __init__(self, *options: Option) -> None:
        """
        Initialise the class with the provided options.

        Examples:
            >>> engin = Engin(Provide(construct_a), Invoke(do_b), Supply(C()), MyBlock())

        Args:
            *options: an instance of Provide, Supply, Invoke, Entrypoint or a Block.
        """
        self._state = _EnginState.IDLE
        self._start_complete_event = Event()
        self._stop_requested_event = Event()
        self._stop_complete_event = Event()
        self._exit_stack = AsyncExitStack()
        self._async_context_run_task: asyncio.Task | None = None

        self._providers: dict[TypeId, Provide] = {}
        self._multiproviders: dict[TypeId, list[Provide]] = defaultdict(list)
        self._invocations: list[Invoke] = []

        # populates the above
        for option in chain(self._LIB_OPTIONS, options):
            option.apply(self)

        # initialise Assembler
        self._assembler = Assembler.from_mapped_providers(
            providers=self._providers,
            multiproviders=self._multiproviders,
        )
        self._assembler.add(Supply(self._assembler))

    @property
    def assembler(self) -> Assembler:
        return self._assembler

    async def run(self) -> None:
        """
        Run the engin.

        The engin will run until it is stopped via an external signal (i.e. SIGTERM or
        SIGINT), the `stop` method is called on the engin, or a lifecycle task errors.
        """
        if self._state != _EnginState.IDLE:
            raise EnginError("Engin is not idle, unable to start")

        LOG.info("starting engin")
        assembled_invocations: list[AssembledDependency] = [
            await self._assembler.assemble(invocation) for invocation in self._invocations
        ]

        for invocation in assembled_invocations:
            try:
                await invocation()
            except Exception as err:
                name = invocation.dependency.name
                LOG.error(f"invocation '{name}' errored, exiting", exc_info=err)
                raise

        lifecycle = await self._assembler.build(Lifecycle)

        try:
            for hook in lifecycle.list():
                await asyncio.wait_for(self._exit_stack.enter_async_context(hook), timeout=15)
        except Exception as err:
            if isinstance(err, TimeoutError):
                msg = "lifecycle startup task timed out after 15s, exiting"
            else:
                msg = "lifecycle startup task errored, exiting"
            LOG.error(msg, exc_info=err)
            await self._shutdown()
            return

        supervisor = await self._assembler.build(Supervisor)

        LOG.info("startup complete")
        self._state = _EnginState.RUNNING
        self._start_complete_event.set()

        async with create_task_group() as tg:
            if self._STOP_ON_SINGAL:
                tg.start_soon(_stop_engin_on_signal, self._stop_requested_event)

            try:
                async with supervisor:
                    await self._stop_requested_event.wait()

                # shutdown after stopping supervised tasks
                await self._shutdown()
            except BaseException:
                await self._shutdown()

            tg.cancel_scope.cancel()

    async def start(self) -> None:
        """
        Starts the engin in the background. This method will wait until the engin is fully
        started to return so it is safe to use immediately after.
        """
        self._async_context_run_task = asyncio.create_task(self.run())
        wait_tasks = [
            asyncio.create_task(self._start_complete_event.wait()),
            asyncio.create_task(self._stop_complete_event.wait()),
        ]
        await asyncio.wait(
            [
                self._async_context_run_task,  # if a provider errors this will return first
                *wait_tasks,
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in wait_tasks:
            task.cancel()

        # raise the exception from the startup during run
        if self._async_context_run_task.done():
            startup_exception = self._async_context_run_task.exception()
            if startup_exception is not None:
                raise startup_exception

    async def stop(self) -> None:
        """
        Stop the engin.

        This method will wait for the shutdown lifecycle to complete before returning.
        Note this method can be safely called at any point, even before the engin is
        started.
        """
        self._stop_requested_event.set()
        if self._state == _EnginState.RUNNING:
            await self._stop_complete_event.wait()

    def graph(self) -> list[Node]:
        """
        Creates a graph representation of the engin's dependencies which can be used for
        introspection or visualisations.

        Returns: a list of Node objects.
        """
        grapher = DependencyGrapher({**self._providers, **self._multiproviders})
        return grapher.resolve(self._invocations)

    def is_running(self) -> bool:
        return self._state == _EnginState.RUNNING

    def is_stopped(self) -> bool:
        return self._state == _EnginState.SHUTDOWN

    async def _shutdown(self) -> None:
        if self._state == _EnginState.RUNNING:
            LOG.info("stopping engin")
            await self._exit_stack.aclose()
            self._stop_complete_event.set()
            LOG.info("shutdown complete")
            self._state = _EnginState.SHUTDOWN


async def _stop_engin_on_signal(stop_requested_event: Event) -> None:
    """
    A task that waits for a stop signal (SIGINT/SIGTERM) and notifies the given event.

    On unix-like systems we can use asyncio's `loop.add_signal_handler` method but this
    does not work on Windows as `signal.set_wakeup_fd` is not supported.

    Therefore on Windows we fallback to using `signal.signal` directly.
    """
    if not _OS_IS_WINDOWS:
        with open_signal_receiver(signal.SIGINT, signal.SIGTERM) as received_signals:
            async for signum in received_signals:
                LOG.debug(f"received signal: {signum.name}")
                stop_requested_event.set()
                break
    else:

        def ctrlc_handler(sig: int, frame: FrameType | None) -> None:
            LOG.debug(f"received signal: {signal.Signals(sig).name}")
            if stop_requested_event.is_set():
                raise KeyboardInterrupt("Forced keyboard interrupt")
            else:
                stop_requested_event.set()

        previous_signal_handler_sigint = signal.signal(signal.SIGINT, ctrlc_handler)

        # technically not needed due to the _OS_IS_WINDOWS checks but keeps mypy happy
        if hasattr(signal, "SIGBREAK"):
            previous_signal_handler_sigbreak = signal.signal(signal.SIGBREAK, ctrlc_handler)

        try:
            await stop_requested_event.wait()
        finally:
            # restore orginal signal handlers
            signal.signal(signal.SIGINT, previous_signal_handler_sigint)
            if hasattr(signal, "SIGBREAK"):
                signal.signal(signal.SIGBREAK, previous_signal_handler_sigbreak)



import inspect
import typing
from collections.abc import Iterable
from inspect import Parameter
from typing import ClassVar, TypeVar

from fastapi.routing import APIRoute

from engin import Assembler, Engin, Entrypoint, Invoke, Option
from engin._dependency import Dependency, Supply, _noop
from engin._graph import DependencyGrapher, Node
from engin._type_utils import TypeId
from engin.extensions.asgi import ASGIEngin

try:
    from fastapi import APIRouter, FastAPI
    from fastapi.params import Depends
    from starlette.requests import HTTPConnection
except ImportError as err:
    raise ImportError(
        "fastapi package must be installed to use the fastapi extension"
    ) from err

__all__ = ["APIRouteDependency", "FastAPIEngin", "Inject"]


def _attach_assembler(app: FastAPI, assembler: Assembler) -> None:
    """
    An invocation that attaches the Engin's Assembler to the FastAPI application, enabling
    the Inject marker.
    """
    app.state.assembler = assembler


class FastAPIEngin(ASGIEngin):
    _LIB_OPTIONS: ClassVar[list[Option]] = [
        *ASGIEngin._LIB_OPTIONS,
        Invoke(_attach_assembler),
    ]
    _asgi_type = FastAPI

    def graph(self) -> list[Node]:
        grapher = _FastAPIDependencyGrapher({**self._providers, **self._multiproviders})
        return grapher.resolve(
            [
                Entrypoint(self._asgi_type),
                *[i for i in self._invocations if i.func_name != "_attach_assembler"],
            ]
        )


T = TypeVar("T")


def Inject(interface: type[T]) -> Depends:
    # note: the APIRouteDependency class looks for this function's name
    async def __inner_engin_dependency__(conn: HTTPConnection) -> T:
        try:
            assembler: Assembler = conn.app.state.assembler
        except AttributeError:
            raise RuntimeError("Assembler is not attached to Application state") from None
        return await assembler.build(interface)

    return Depends(__inner_engin_dependency__)


class _FastAPIDependencyGrapher(DependencyGrapher):
    """
    This exists in order to bridge the gap between
    """

    def _resolve_recursive(
        self, roots: Iterable[Dependency], *, seen: set[TypeId]
    ) -> list[Node]:
        nodes: list[Node] = []
        for root in roots:
            for parameter in root.parameter_type_ids:
                provider = self._providers[parameter]

                # multiprovider
                if isinstance(provider, list):
                    for p in provider:
                        nodes.append(Node(node=p, parent=root))

                        if isinstance(p, Supply):
                            route_dependencies = _extract_routes_from_supply(p)
                            nodes.extend(
                                Node(node=route_dependency, parent=p)
                                for route_dependency in route_dependencies
                            )
                            nodes.extend(
                                self._resolve_recursive(route_dependencies, seen=seen)
                            )

                        if parameter not in seen:
                            nodes.extend(self._resolve_recursive(provider, seen=seen))
                # single provider
                else:
                    nodes.append(Node(node=provider, parent=root))
                    # not sure why anyone would ever supply a single APIRouter in an
                    # application, but just in case
                    if isinstance(provider, Supply):
                        route_dependencies = _extract_routes_from_supply(provider)
                        nodes.extend(
                            Node(node=route_dependency, parent=provider)
                            for route_dependency in route_dependencies
                        )
                        nodes.extend(self._resolve_recursive(route_dependencies, seen=seen))
                    if parameter not in seen:
                        nodes.extend(self._resolve_recursive([provider], seen=seen))

                seen.add(parameter)

        return nodes


def _extract_routes_from_supply(supply: Supply) -> list[Dependency]:
    if supply.is_multiprovider:
        inner = supply._value[0]
        if isinstance(inner, APIRouter):
            return [
                APIRouteDependency(supply, route)
                for route in inner.routes
                if isinstance(route, APIRoute)
            ]
    return []


class APIRouteDependency(Dependency):
    """
    This is a pseudo-dependency that is only used when calling FastAPIEngin.graph() in
    order to provide richer metadata to the Node.

    This class should never be constructed in application code.
    """

    def __init__(
        self,
        wraps: Dependency,
        route: APIRoute,
    ) -> None:
        """
        Warning: this should never be constructed in application code.
        """
        super().__init__(_noop)
        self._block_name = wraps.block_name
        self._wrapped = wraps
        self._route = route
        self._signature = inspect.signature(route.endpoint)

    @property
    def source_module(self) -> str:
        return self._wrapped.source_module

    @property
    def source_package(self) -> str:
        return self._wrapped.source_package

    @property
    def route(self) -> APIRoute:
        return self._route

    @property
    def parameter_type_ids(self) -> list[TypeId]:
        parameters = list(self._signature.parameters.values())
        if not parameters:
            return []
        if parameters[0].name == "self":
            parameters.pop(0)
        return [
            TypeId.from_type(typing.get_args(param.annotation)[0])
            for param in parameters
            if self._is_injected_param(param)
        ]

    @staticmethod
    def _is_injected_param(param: Parameter) -> bool:
        if typing.get_origin(param.annotation) != typing.Annotated:
            return False
        args = typing.get_args(param.annotation)
        if len(args) != 2:
            return False
        return (
            isinstance(args[1], Depends)
            and (depenency_function := getattr(args[1], "dependency", None)) is not None
            and inspect.isfunction(depenency_function)
            and depenency_function.__name__ == "__inner_engin_dependency__"
        )

    @property
    def name(self) -> str:
        methods = ",".join(self._route.methods)
        return f"{methods} {self._route.path}"

    def apply(self, engin: Engin) -> None:
        raise NotImplementedError("APIRouteDependency is not a real dependency")



from typing import Annotated

import typer
from rich import box
from rich.console import Console
from rich.table import Table

from engin import Supply
from engin._cli._common import COMMON_HELP, get_engin_instance, print_error

cli = typer.Typer()
_CLI_HELP = {
    "type": "Filter providers by the provided type, e.g. `AsyncClient` or `float[]`",
    "module": "Filter providers by the provided type's module, e.g. `engin` or `httpx`",
    "verbose": "Enables verbose output",
}


@cli.command(name="inspect")
def serve_graph(
    app: Annotated[
        str | None,
        typer.Argument(help=COMMON_HELP["app"]),
    ] = None,
    type_: Annotated[
        str | None,
        typer.Option("--type", help=_CLI_HELP["type"]),
    ] = None,
    module: Annotated[
        str | None,
        typer.Option(help=_CLI_HELP["module"]),
    ] = None,
    verbose: Annotated[
        bool, typer.Option("--verbose", "-v", help=_CLI_HELP["verbose"])
    ] = False,
) -> None:
    """
    Shows metadata for all matching providers.

    Examples:

        1. `engin inspect --module httpx`
        2. `engin inspect --type AsyncClient`
    """
    module_name, _, instance = get_engin_instance(app)

    console = Console()

    providers = []
    for provider in instance.assembler.providers:
        type_id = provider.return_type_id
        if type_ is not None:
            type_name = str(type_id).rsplit(".", maxsplit=1)[-1]
            if type_ != type_name:
                if verbose:
                    console.print(
                        f"Ignoring '{provider.return_type_id}' as `{type_} != {type_name}",
                        style="dim",
                    )
                continue
        if module is not None:
            module_name = str(type_id).split(".", maxsplit=1)[0]
            if module != module_name:
                if verbose:
                    console.print(
                        f"Ignoring '{provider.return_type_id}' as `{module} != {module_name}",
                        style="dim",
                    )
                continue
        providers.append(provider)

    matching_provider_count = len(providers)
    if matching_provider_count == 0:
        available = sorted(map(str, instance.assembler.providers))
        print_error(f"No matching providers, available: {available}")

    console.print(
        f"Found {matching_provider_count} matching provider"
        + ("s" if matching_provider_count > 1 else ""),
        style="dim",
    )

    table = Table(show_header=False, show_lines=False, box=box.ASCII)

    for provider in sorted(providers, key=lambda p: p.source_module):
        is_supply = isinstance(provider, Supply)

        table.add_row("name", str(provider), style="bold", end_section=True)
        table.add_row("scope", provider.scope or "N/A")
        table.add_row("func", provider.func_name if not is_supply else "N/A")
        table.add_row("block", provider.block_name or "N/A")
        table.add_row("source module", provider.source_module or "N/A")
        table.add_row("source package", provider.source_package or "N/A")
        table.add_section()

    console.print(table)



import logging
import sys

try:
    import typer
except ImportError:
    raise ImportError(
        "Unable to import typer, to use the engin cli please install the"
        " `cli` extra, e.g. pip install engin[cli]"
    ) from None

from engin._cli._check import cli as check_cli
from engin._cli._graph import cli as graph_cli
from engin._cli._inspect import cli as inspect_cli

# mute logging from importing of files + engin's debug logging.
logging.disable()

# add cwd to path to enable local package imports
sys.path.insert(0, "")

app = typer.Typer()

app.add_typer(check_cli)
app.add_typer(graph_cli)
app.add_typer(inspect_cli)



from abc import abstractmethod
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from engin._engin import Engin


class Option(Protocol):
    @abstractmethod
    def apply(self, engin: "Engin") -> None: ...
